@@ -69,9 +69,16 @@
 #ifndef EV_SYN
 #define EV_SYN 0
 #endif
+#ifndef SYN_MAX
+#define SYN_MAX 3
+#define SYN_CNT (SYN_MAX + 1)
+#endif
 #ifndef SYN_MT_REPORT
 #define SYN_MT_REPORT 2
 #endif
+#ifndef SYN_DROPPED
+#define SYN_DROPPED 3
+#endif
 
 #define NAME_ELEMENT(element) [element] = #element
 
@@ -162,6 +169,32 @@ static const char * const events[EV_MAX + 1] = {
 	NAME_ELEMENT(EV_FF_STATUS),		NAME_ELEMENT(EV_SW),
 };
 
+static const int maxval[EV_MAX + 1] = {
+	[0 ... EV_MAX] = -1,
+	[EV_SYN] = SYN_MAX,
+	[EV_KEY] = KEY_MAX,
+	[EV_REL] = REL_MAX,
+	[EV_ABS] = ABS_MAX,
+	[EV_MSC] = MSC_MAX,
+	[EV_SW] = SW_MAX,
+	[EV_LED] = LED_MAX,
+	[EV_SND] = SND_MAX,
+	[EV_REP] = REP_MAX,
+	[EV_FF] = FF_MAX,
+	[EV_FF_STATUS] = FF_STATUS_MAX,
+};
+
+
+#ifdef INPUT_PROP_SEMI_MT
+static const char * const props[INPUT_PROP_MAX + 1] = {
+	[0 ... INPUT_PROP_MAX] = NULL,
+	NAME_ELEMENT(INPUT_PROP_POINTER),
+	NAME_ELEMENT(INPUT_PROP_DIRECT),
+	NAME_ELEMENT(INPUT_PROP_BUTTONPAD),
+	NAME_ELEMENT(INPUT_PROP_SEMI_MT),
+};
+#endif
+
 static const char * const keys[KEY_MAX + 1] = {
 	[0 ... KEY_MAX] = NULL,
 	NAME_ELEMENT(KEY_RESERVED),		NAME_ELEMENT(KEY_ESC),
@@ -368,6 +401,15 @@ static const char * const keys[KEY_MAX + 1] = {
 	NAME_ELEMENT(KEY_TOUCHPAD_ON),
 	NAME_ELEMENT(KEY_TOUCHPAD_OFF),
 #endif
+#ifdef KEY_CAMERA_ZOOMIN
+	NAME_ELEMENT(KEY_CAMERA_ZOOMIN),	NAME_ELEMENT(KEY_CAMERA_ZOOMOUT),
+	NAME_ELEMENT(KEY_CAMERA_UP),		NAME_ELEMENT(KEY_CAMERA_DOWN),
+	NAME_ELEMENT(KEY_CAMERA_LEFT),		NAME_ELEMENT(KEY_CAMERA_RIGHT),
+#endif
+#ifdef KEY_ATTENDANT_ON
+	NAME_ELEMENT(KEY_ATTENDANT_ON),		NAME_ELEMENT(KEY_ATTENDANT_OFF),
+	NAME_ELEMENT(KEY_ATTENDANT_TOGGLE),	NAME_ELEMENT(KEY_LIGHTS_TOGGLE),
+#endif
 
 	NAME_ELEMENT(BTN_0),			NAME_ELEMENT(BTN_1),
 	NAME_ELEMENT(BTN_2),			NAME_ELEMENT(BTN_3),
@@ -384,9 +426,14 @@ static const char * const keys[KEY_MAX + 1] = {
 	NAME_ELEMENT(BTN_BASE),			NAME_ELEMENT(BTN_BASE2),
 	NAME_ELEMENT(BTN_BASE3),		NAME_ELEMENT(BTN_BASE4),
 	NAME_ELEMENT(BTN_BASE5),		NAME_ELEMENT(BTN_BASE6),
-	NAME_ELEMENT(BTN_DEAD),			NAME_ELEMENT(BTN_A),
-	NAME_ELEMENT(BTN_B),			NAME_ELEMENT(BTN_C),
+	NAME_ELEMENT(BTN_DEAD),			NAME_ELEMENT(BTN_C),
+#ifdef BTN_SOUTH
+	NAME_ELEMENT(BTN_SOUTH),		NAME_ELEMENT(BTN_EAST),
+	NAME_ELEMENT(BTN_NORTH),		NAME_ELEMENT(BTN_WEST),
+#else
+	NAME_ELEMENT(BTN_A),			NAME_ELEMENT(BTN_B),
 	NAME_ELEMENT(BTN_X),			NAME_ELEMENT(BTN_Y),
+#endif
 	NAME_ELEMENT(BTN_Z),			NAME_ELEMENT(BTN_TL),
 	NAME_ELEMENT(BTN_TR),			NAME_ELEMENT(BTN_TL2),
 	NAME_ELEMENT(BTN_TR2),			NAME_ELEMENT(BTN_SELECT),
@@ -402,6 +449,11 @@ static const char * const keys[KEY_MAX + 1] = {
 	NAME_ELEMENT(BTN_GEAR_DOWN),
 	NAME_ELEMENT(BTN_GEAR_UP),
 
+#ifdef BTN_DPAD_UP
+	NAME_ELEMENT(BTN_DPAD_UP),		NAME_ELEMENT(BTN_DPAD_DOWN),
+	NAME_ELEMENT(BTN_DPAD_LEFT),		NAME_ELEMENT(BTN_DPAD_RIGHT),
+#endif
+
 #ifdef BTN_TRIGGER_HAPPY
 	NAME_ELEMENT(BTN_TRIGGER_HAPPY1),	NAME_ELEMENT(BTN_TRIGGER_HAPPY11),
 	NAME_ELEMENT(BTN_TRIGGER_HAPPY2),	NAME_ELEMENT(BTN_TRIGGER_HAPPY12),
@@ -425,6 +477,9 @@ static const char * const keys[KEY_MAX + 1] = {
 	NAME_ELEMENT(BTN_TRIGGER_HAPPY29),	NAME_ELEMENT(BTN_TRIGGER_HAPPY39),
 	NAME_ELEMENT(BTN_TRIGGER_HAPPY30),	NAME_ELEMENT(BTN_TRIGGER_HAPPY40),
 #endif
+#ifdef BTN_TOOL_QUINTTAP
+	NAME_ELEMENT(BTN_TOOL_QUINTTAP),
+#endif
 };
 
 static const char * const absval[6] = { "Value", "Min  ", "Max  ", "Fuzz ", "Flat ", "Resolution "};
@@ -474,6 +529,11 @@ static const char * const absolutes[ABS_MAX + 1] = {
 #ifdef ABS_MT_SLOT
 	NAME_ELEMENT(ABS_MT_SLOT),
 #endif
+#ifdef ABS_MT_TOOL_X
+	NAME_ELEMENT(ABS_MT_TOOL_X),
+	NAME_ELEMENT(ABS_MT_TOOL_Y),
+	NAME_ELEMENT(ABS_MT_DISTANCE),
+#endif
 
 };
 
@@ -482,6 +542,9 @@ static const char * const misc[MSC_MAX + 1] = {
 	NAME_ELEMENT(MSC_SERIAL),		NAME_ELEMENT(MSC_PULSELED),
 	NAME_ELEMENT(MSC_GESTURE),		NAME_ELEMENT(MSC_RAW),
 	NAME_ELEMENT(MSC_SCAN),
+#ifdef MSC_TIMESTAMP
+	NAME_ELEMENT(MSC_TIMESTAMP),
+#endif
 };
 
 static const char * const leds[LED_MAX + 1] = {
@@ -504,12 +567,12 @@ static const char * const sounds[SND_MAX + 1] = {
 	NAME_ELEMENT(SND_TONE)
 };
 
-static const char * const syns[3] = {
+static const char * const syns[SYN_MAX + 1] = {
+	[0 ... SYN_MAX] = NULL,
 	NAME_ELEMENT(SYN_REPORT),
 	NAME_ELEMENT(SYN_CONFIG),
-#ifdef SYN_MT_REPORT
-	NAME_ELEMENT(SYN_MT_REPORT)
-#endif
+	NAME_ELEMENT(SYN_MT_REPORT),
+	NAME_ELEMENT(SYN_DROPPED)
 };
 
 static const char * const switches[SW_MAX + 1] = {
@@ -622,7 +685,7 @@ static char* scan_devices(void)
 	int i, ndev, devnum;
 	char *filename;
 
-	ndev = scandir(DEV_INPUT_EVENT, &namelist, is_event_device, alphasort);
+	ndev = scandir(DEV_INPUT_EVENT, &namelist, is_event_device, versionsort);
 	if (ndev <= 0)
 		return NULL;
 
@@ -708,6 +771,37 @@ static void print_absdata(int fd, int axis)
 			printf("      %s %6d\n", absval[k], abs[k]);
 }
 
+static void print_repdata(int fd)
+{
+	int i;
+	unsigned int rep[2];
+
+	ioctl(fd, EVIOCGREP, rep);
+
+	for (i = 0; i <= REP_MAX; i++) {
+		printf("    Repeat code %d (%s)\n", i, names[EV_REP] ? (names[EV_REP][i] ? names[EV_REP][i] : "?") : "?");
+		printf("      Value %6d\n", rep[i]);
+	}
+
+}
+
+static inline const char* typename(unsigned int type)
+{
+	return (type <= EV_MAX && events[type]) ? events[type] : "?";
+}
+
+static inline const char* codename(unsigned int type, unsigned int code)
+{
+	return (type <= EV_MAX && code <= maxval[type] && names[type] && names[type][code]) ? names[type][code] : "?";
+}
+
+#ifdef INPUT_PROP_SEMI_MT
+static inline const char* propname(unsigned int prop)
+{
+	return (prop <= INPUT_PROP_MAX && props[prop]) ? props[prop] : "?";
+}
+#endif
+
 /**
  * Print static device information (no events). This information includes
  * version numbers, device name and all bits supported by this device.
@@ -717,11 +811,14 @@ static void print_absdata(int fd, int axis)
  */
 static int print_device_info(int fd)
 {
-	int i, j;
+	unsigned int type, code, prop;
 	int version;
 	unsigned short id[4];
 	char name[256] = "Unknown";
 	unsigned long bit[EV_MAX][NBITS(KEY_MAX)];
+#ifdef INPUT_PROP_SEMI_MT
+	unsigned long propbits[INPUT_PROP_MAX];
+#endif
 
 	if (ioctl(fd, EVIOCGVERSION, &version)) {
 		perror("evtest: can't get version");
@@ -742,18 +839,34 @@ static int print_device_info(int fd)
 	ioctl(fd, EVIOCGBIT(0, EV_MAX), bit[0]);
 	printf("Supported events:\n");
 
-	for (i = 0; i < EV_MAX; i++)
-		if (test_bit(i, bit[0])) {
-			printf("  Event type %d (%s)\n", i, events[i] ? events[i] : "?");
-			if (!i) continue;
-			ioctl(fd, EVIOCGBIT(i, KEY_MAX), bit[i]);
-			for (j = 0; j < KEY_MAX; j++)
-				if (test_bit(j, bit[i])) {
-					printf("    Event code %d (%s)\n", j, names[i] ? (names[i][j] ? names[i][j] : "?") : "?");
-					if (i == EV_ABS)
-						print_absdata(fd, j);
+	for (type = 0; type < EV_MAX; type++) {
+		if (test_bit(type, bit[0]) && type != EV_REP) {
+			printf("  Event type %d (%s)\n", type, typename(type));
+			if (type == EV_SYN) continue;
+			ioctl(fd, EVIOCGBIT(type, KEY_MAX), bit[type]);
+			for (code = 0; code < KEY_MAX; code++)
+				if (test_bit(code, bit[type])) {
+					printf("    Event code %d (%s)\n", code, codename(type, code));
+					if (type == EV_ABS)
+						print_absdata(fd, code);
 				}
 		}
+	}
+
+	if (test_bit(EV_REP, bit[0])) {
+		printf("Key repeat handling:\n");
+		printf("  Repeat type %d (%s)\n", EV_REP, events[EV_REP] ?  events[EV_REP] : "?");
+		print_repdata(fd);
+	}
+#ifdef INPUT_PROP_SEMI_MT
+	memset(propbits, 0, sizeof(propbits));
+	ioctl(fd, EVIOCGPROP(sizeof(propbits)), propbits);
+	printf("Properties:\n");
+	for (prop = 0; prop < INPUT_PROP_MAX; prop++) {
+		if (test_bit(prop, propbits))
+			printf("  Property type %d (%s)\n", prop, propname(prop));
+	}
+#endif
 
 	return 0;
 }
@@ -779,20 +892,23 @@ static int print_events(int fd)
 		}
 
 		for (i = 0; i < rd / sizeof(struct input_event); i++) {
+			unsigned int type, code;
+
+			type = ev[i].type;
+			code = ev[i].code;
+
 			printf("Event: time %ld.%06ld, ", ev[i].time.tv_sec, ev[i].time.tv_usec);
 
-			if (ev[i].type == EV_SYN) {
-				if (ev[i].code == SYN_MT_REPORT)
-					printf("++++++++++++++ %s ++++++++++++\n", syns[ev[i].code]);
+			if (type == EV_SYN) {
+				if (code == SYN_MT_REPORT)
+					printf("++++++++++++++ %s ++++++++++++\n", codename(type, code));
 				else
-					printf("-------------- %s ------------\n", syns[ev[i].code]);
+					printf("-------------- %s ------------\n", codename(type, code));
 			} else {
 				printf("type %d (%s), code %d (%s), ",
-					ev[i].type,
-					events[ev[i].type] ? events[ev[i].type] : "?",
-					ev[i].code,
-					names[ev[i].type] ? (names[ev[i].type][ev[i].code] ? names[ev[i].type][ev[i].code] : "?") : "?");
-				if (ev[i].type == EV_MSC && (ev[i].code == MSC_RAW || ev[i].code == MSC_SCAN))
+					type, typename(type),
+					code, codename(type, code));
+				if (type == EV_MSC && (code == MSC_RAW || code == MSC_SCAN))
 					printf("value %02x\n", ev[i].value);
 				else
 					printf("value %d\n", ev[i].value);
@@ -830,7 +946,7 @@ static int test_grab(int fd)
 static int do_capture(const char *device)
 {
 	int fd;
-	char *filename;
+	char *filename = NULL;
 
 	if (!device) {
 		fprintf(stderr, "No device specified, trying to scan all of %s/%s*\n",
@@ -854,16 +970,14 @@ static int do_capture(const char *device)
 			fprintf(stderr, "You do not have access to %s. Try "
 					"running as root instead.\n",
 					filename);
-		return EXIT_FAILURE;
+		goto error;
 	}
 
-	free(filename);
-
 	if (!isatty(fileno(stdout)))
 		setbuf(stdout, NULL);
 
 	if (print_device_info(fd))
-		return EXIT_FAILURE;
+		goto error;
 
 	printf("Testing ... (interrupt to exit)\n");
 
@@ -875,10 +989,19 @@ static int do_capture(const char *device)
 		       "  other grab is active.\n");
 		printf("  In most cases, this is caused by an X driver,\n"
 		       "  try VT-switching and re-run evtest again.\n");
+		printf("  Run the following command to see processes with\n"
+		       "  an open fd on this device\n"
+		       " \"fuser -v %s\"\n", filename);
 		printf("***********************************************\n");
 	}
 
+	free(filename);
+
 	return print_events(fd);
+
+error:
+	free(filename);
+	return EXIT_FAILURE;
 }
 
 /**
